from win32com.client import constants
from pywintypes import com_error
import time
import traceback, sys

from .wait import _wait_for_capiq_result
from ..exceptions import WorkbookClosedException, CapitalIQInactiveException
from exceldriver.tools import _restart_excel_with_addins_and_attach

def populate_capiq_for_file(filepath, excel, retries_remaining=3, close_workbook=False, index=0):
    """

    Private function has main functionality. This is a wrapper to add retries afer com errors
    """

    # Even if things are going normally, restart every 500 worksheets as there is a memory leak
    if index % 500 == 0 and retries_remaining == 3:
        excel = _restart_excel_with_addins_and_attach()

    # Stop retries
    if retries_remaining <= 0:
        print(fr'ERROR: Could not process {filepath}. Skipping and moving to "..\failed".')
        return excel, False

    try:
        # If we are retrying, need to close the workbook before trying to populate
        if close_workbook:
            excel.CutCopyMode = False
            time.sleep(1)
            if excel.ActiveWorkbook:
                excel.ActiveWorkbook.Close(SaveChanges=False)
            time.sleep(5)

        _populate_capiq_for_file(filepath, excel)
        return excel, True
    except (com_error, WorkbookClosedException, CapitalIQInactiveException) as e:
        print(f'Error {e} populating {filepath}. Will wait 30 seconds, restart Excel, and try again.')
        traceback.print_tb(sys.exc_info()[2])
        time.sleep(30)
        excel = _restart_excel_with_addins_and_attach()
        return populate_capiq_for_file(filepath, excel, retries_remaining=retries_remaining - 1, close_workbook=True)

def _populate_capiq_for_file(filepath, excel):
    wb = excel.Workbooks.Open(filepath)
    successful = _wait_for_capiq_result(excel)
    _copy_paste_values(excel, wb)
    excel.ActiveWorkbook.Close(SaveChanges=True)
    return successful

def populate_capiq_ids_for_file(filepath, excel):
    wb = excel.Workbooks.Open(filepath)
    successful = _wait_for_capiq_result(excel)
    _copy_paste_values(excel, wb, range='A1:H3000')
    excel.ActiveWorkbook.Close(SaveChanges=True)
    return successful

def _copy_paste_values(excel, wb, range='A1:ZZ20000'):
    ws = wb.Sheets('Sheet')
    ws.Range(range).Copy()
    ws.Range(range.split(':')[0]).PasteSpecial(Paste=constants.xlPasteValues, Operation=constants.xlNone)
    excel.CutCopyMode = False


import multiprocessing as mp

class TimeoutWrapper:

    def __init__(self, timeout, func, timeout_callback):
        self.timeout = timeout
        self.func = func
        self.timeout_callback = timeout_callback

    def __call__(self,  *func_args, timeout_callback_kwargs={},  **func_kwargs):

        with mp.Pool(1) as p:
            result_promise = p.apply_async(self.func, args=func_args, kwds=func_kwargs, error_callback=_raise_exception)

            try:
                result = result_promise.get(self.timeout)
                return result
            except mp.TimeoutError:
                # Run timeout callback outside of pool
                pass

        # code only reaches here in mp.TimeoutError
        return self.timeout_callback(**timeout_callback_kwargs)



def _raise_exception(exception: Exception):
    raise exception



def financial_data_command(company_id, data_item, freq='Q', num_periods=80, data_item_label=None):
    _validate_financial_data_inputs(company_id, data_item, freq=freq, num_periods=num_periods, data_item_label=data_item_label)

    # Use variable name as label if none specified
    if data_item_label is None:
        data_item_label = data_item

    return f'=CIQRANGE("{company_id}", "{data_item}", IQ_F{freq} - {num_periods}, , , , , , "{data_item_label}")'

def holdings_command(company_id, data_item, date_str, data_item_label=None):
    # Use variable name as label if none specified
    if data_item_label is None:
        data_item_label = data_item

    return f'=CIQRANGE("{company_id}", "{data_item}", 1, 50000, "{date_str}", , , , "{data_item_label}")'

def id_command(search_str):
    return f'=CIQRANGEA("{search_str}","IQ_COMPANY_ID_QUICK_MATCH",1,1)'

def name_command(search_str):
    return f'=CIQRANGEA("{search_str}","IQ_COMPANY_NAME_QUICK_MATCH",1,1)'

def _validate_financial_data_inputs(*args, **kwargs):
    assert kwargs['freq'] in ('Q','Y')
    assert isinstance(args[0], str)
    assert isinstance(args[1], str)


import pandas as pd
import shutil
import os
import time
import re
import datetime
from functools import partial

def append_csvs_to_csv(csv_filepath_list, outpath=None):
    """
    Appends csvs into a single csv. Is memory efficient by only keeping the current
    processed file in memory. However still keeps track of changing columns to ensure
    data is correctly aligned.

    :param csv_filepath_list:
    :param outpath:
    :return:
    """

    outpath, df_of_headers = _get_outpath_and_df_of_headers(outpath)

    all_columns = [col for col in df_of_headers.columns]

    for file in csv_filepath_list:
        df_for_append = pd.read_csv(file) #load new data
        df_of_headers, all_columns = _append_df_to_csv(df_for_append, df_of_headers, outpath, all_columns)

def append_csv_to_csv(inpath, outpath):
    return append_csvs_to_csv([inpath], outpath)

def append_csvs_to_monthly_csv_of_first_date(csv_filepath_list, rootname):
    [append_csv_to_monthly_csv_of_first_date(inpath, rootname) for inpath in csv_filepath_list]

def append_csv_to_monthly_csv_of_first_date(inpath, rootname):
    df_for_append = pd.read_csv(inpath)  # load new data
    return append_df_to_monthly_csv_of_first_date(df_for_append, rootname)

def append_df_to_monthly_csv_of_first_date(df_for_append, rootname):
    year_month_str = first_year_month_in_df(df_for_append)

    outpath = rootname + year_month_str + '.csv'
    append_df_to_csv(df_for_append, outpath)

def append_df_to_csv(df, outpath):
    outpath, df_of_headers = _get_outpath_and_df_of_headers(outpath)
    all_columns = [col for col in df_of_headers.columns]
    _append_df_to_csv(df, df_of_headers, outpath, all_columns)

def _get_outpath_and_df_of_headers(outpath):
    # Output to a new csv if outpath not supplied
    if outpath is None:
        outpath = 'combined.csv'
    # If outpath doesn't exists, must create df from scratch
    if (outpath is None) or (not os.path.exists(outpath)):
        df_of_headers = pd.DataFrame()
    # If exists, load headers
    else:
        df_of_headers = pd.read_csv(outpath, nrows=0)

    return outpath, df_of_headers

def _append_df_to_csv(df_for_append, df_of_headers, outpath, all_columns):

    # Set mode to append or write
    # If there are any columns from loaded csv, then don't write headers. If there are none, write.
    if not all_columns:
        header = True
    else:
        header = False

    # Get headers for combined data
    new_columns = [col for col in df_for_append.columns if col not in df_of_headers.columns]
    all_columns += new_columns
    df_of_headers = pd.DataFrame(columns=all_columns)

    # Append new data, with or without writing headers to file
    full_df = df_of_headers.append(df_for_append)
    full_df[all_columns].to_csv(outpath, mode='a', index=False, header=header, encoding='utf8')

    # Go back to file and update the headers
    headers_csv = pd.DataFrame(columns=all_columns).to_csv(index=False)
    replace_first_line_of_file(outpath, headers_csv)

    return df_of_headers, all_columns


def replace_first_line_of_file(inpath, new_first_line, outpath=None):

    if outpath is None:
        rename = True
        outpath = inpath + '.temp' #replace existing file
    else:
        rename = False

    with open(inpath, 'r', encoding='utf8') as infile:
        infile.readline()  # and discard
        with open(outpath, 'w', encoding='utf8') as outfile:
            outfile.write(new_first_line)
            shutil.copyfileobj(infile, outfile) #bring over rest of contents to new (or replaced) file

    if rename:
        _replace(outpath, inpath)


def _replace(src, dst):
    retries = 0
    while True:
        try:
            os.replace(src, dst)
            break
        except (PermissionError, OSError) as e:
            time.sleep(.1)
            retries += 1
            if retries > 100:
                print(f'Retried removing {filepath} over 10s but still failed.')
                raise e


ymd_pattern = re.compile(r'(\d\d\d\d)-(\d\d)-(\d\d)')

def first_year_month_in_df(df, date_str_col='Date'):
    return _year_month_from_date_string(df.loc[0, date_str_col])

def _year_month_from_date_string(date_string):
    m = ymd_pattern.match(date_string)

    return f'{m.group(1)}{m.group(2)}'

def add_year_month_column_from_date_string_column(df, date_str_col='Date'):
    """
    note: inplace
    """

    df['YearMonth'] = df[date_str_col].apply(_year_month_from_date_string)


def convert_excel_date_to_pandas_date(exceldates):
    epoch = datetime.datetime(1899, 12, 30)
    convert = partial(_convert_excel_date_to_pandas_date, epoch)

    return exceldates.apply(convert)


def _convert_excel_date_to_pandas_date(epoch, date):
    if not pd.isnull(date) and isinstance(date, (int, float)):
        return epoch + datetime.timedelta(days=date)
    else:
        return date


def scale_variables(df, rescale_factor, rescale_vars):
    """
    Note: inplace
    """
    for var in rescale_vars:
        df[var] = df[var] * rescale_factor

def _how_merge_df(df: pd.DataFrame, other_df: pd.DataFrame, ids, how='left'):
    return df.merge(other_df, on=ids, how=how)

def outer_merge_df(df: pd.DataFrame, other_df: pd.DataFrame, ids):
    return _how_merge_df(df, other_df, ids, how='outer')

def left_merge_df(df: pd.DataFrame, other_df: pd.DataFrame, ids):
    return _how_merge_df(df, other_df, ids, how='left')

def right_merge_df(df: pd.DataFrame, other_df: pd.DataFrame, ids):
    return _how_merge_df(df, other_df, ids, how='right')

def get_full_date_df_from_date_series(date_series, new_date_name=None):

    if new_date_name is None:
        new_date_name = date_series.name

    start_date = date_series.min()
    end_date = date_series.max()

    dates = pd.date_range(
        start=start_date,
        end=end_date,
        freq='D'
    )

    date_df = pd.DataFrame(dates, columns=[new_date_name])

    return date_df


def date_from_year_month(df: pd.DataFrame, day: int = 15, month_col: str = 'Month',
                         year_col: str = 'Year', date_col: str = 'Date') -> None:
    """
    Note: inplace
    """

    _date_str = partial(
        _date_str_from_year_month,
        day=day,
        month_col=month_col,
        year_col=year_col
    )

    df[date_col] = df.apply(_date_str, axis=1)
    df[date_col] = pd.to_datetime(df[date_col])


def _date_str_from_year_month(row_series: pd.Series, day: int = 15, month_col: str = 'Month',
                              year_col: str = 'Year'):
    return f'{int(row_series[month_col])}/{day}/{int(row_series[year_col])}'


import pandas as pd
import warnings

from capiq_excel.downloader.timeout import TimeoutWrapper
from capiq_excel.fileops import get_path_of_failed_folder_add_if_necessary, move_file_to_failed_folder, get_path_of_additional_failed_folder_add_if_necessary
from capiq_excel.workbook.populate import populate_capiq_for_file
from exceldriver.tools import _start_excel_with_addins_and_attach, _get_excel_running_workbook, _restart_excel_with_addins_and_attach, NoExcelWorkbookException
from processfiles.files import FileProcessTracker



def populate_all_files_in_folder(folder, restart=True, timeout=240, run_failed=False):
    """

    """
    _validate_populate_inputs(folder, restart, run_failed)

    excel = _start_excel_with_addins_and_attach()
    failed_folder = get_path_of_failed_folder_add_if_necessary(folder)

    if run_failed:
        # Set main folder as 'failed', then set failed folder as another failed folder inside the original
        folder = failed_folder
        failed_folder = get_path_of_additional_failed_folder_add_if_necessary(folder)

    file_tracker = FileProcessTracker(folder=folder, restart=restart, file_types=('xlsx',))


    populate_all_files_in_folder_with_timeout = TimeoutWrapper(timeout, _populate_capiq_for_multiprocess, timeout_callback=_return_false)


    for i, file in enumerate(file_tracker.file_generator()):
        successful = populate_all_files_in_folder_with_timeout(file, index=i + 1)
        if not successful:
            move_file_to_failed_folder(file, failed_folder)


def _get_company_id_list(id_filepath, id_col='IQID'):
    df = pd.read_csv(id_filepath, usecols=[id_col])
    unique = df[id_col].unique().tolist()

    # Drop nans
    return [i for i in unique if not pd.isnull(i)]

def _validate_populate_inputs(folder, restart, run_failed):
    assert not (restart and run_failed)

    if run_failed:
        warnings.warn(f'run_failed flag passed. Folder {folder}, will not be run, instead the failed folder will')

### Functions below to assist with multiprocessing/timeout handling


def _populate_capiq_for_multiprocess(file, **kwargs):
    try:
        excel = _get_excel_running_workbook('Book1.xlsx')
    except NoExcelWorkbookException:
        excel = _restart_excel_with_addins_and_attach()
    excel, successful = populate_capiq_for_file(file, excel, **kwargs)
    return successful

def _restart_excel_and_return_false():
    excel = _restart_excel_with_addins_and_attach(max_retries=10)

    return False

def _return_false():
    return False

## END TEMP


import pandas as pd

from exceldriver.tools import _start_excel_with_addins_and_attach
from processfiles.files import FileProcessTracker
from .workbook.populate import populate_capiq_ids_for_file


def populate_all_ids_in_folder(folder, restart=True):
    excel = _start_excel_with_addins_and_attach()
    file_tracker = FileProcessTracker(folder=folder, restart=restart, file_types=('xlsx',))

    for file in file_tracker.file_generator():
        populate_capiq_ids_for_file(file, excel)


def combine_all_capiq_ids_xlsx(infolder, outpath, restart=True):

    df = pd.DataFrame()

    file_tracker = FileProcessTracker(folder=infolder, restart=restart, file_types=('xlsx',))

    for file in file_tracker.file_generator():
        df = _append_capiq_xlsx_to_df(df, file)

    _remove_useless_cols(df)
    df.to_csv(outpath, index=False)

    return df


def _append_capiq_xlsx_to_df(df, filepath):
    temp_df = pd.read_excel(filepath)
    df = df.append(temp_df)

    return df


def _remove_useless_cols(df):
    """
    NOTE: inplace
    """
    blank_cols = [col for col in df.columns if 'blank' in col.lower()]
    useless_cols = ['ID'] + blank_cols
    df.drop(useless_cols, axis=1, inplace=True)



from typing import List, Dict
from capiq_excel.workbook.create import (
    create_all_xlsx_with_financials_commands,
    create_all_xlsx_with_holdings_commands,
    create_all_xlsx_with_id_commands
)
from capiq_excel.downloader.tools import populate_all_files_in_folder
from capiq_excel.ids import populate_all_ids_in_folder, combine_all_capiq_ids_xlsx
from capiq_excel.combine import combine_all_capiq_xlsx


def download_financials(folder: str, company_ids: List[str], data_items: Dict[str, str],
                        outpath: str = 'combined.csv',
                        restart: bool = True, timeout: int = 240, run_failed: bool = False,
                        **financial_command_kwargs):
    """
    Downloads financial data from Capital IQ given the capital IQ ids and chosen variables

    :param folder: Location for in-process files
    :param company_ids: Capital IQ ids for companies, e.g. ['IQ4564656', 'IQ45643215']
    :param data_items: Data items to pull for companies. keys are variable names in capital iq, values are what the
        names should be in the output data. e.g. {'IQ_TOTAL_REVENUE': 'Sales'}
    :param outpath: Path to csv where final data should be outputted
    :param restart: If not the first run, start over from the beginning. Set to False to start from where left off
    :param timeout: Time to wait for file to be populated before considering it failed
    :param run_failed: Should only be set to True on a second or later run. Will target the failed files instead
        of the main files if True is passed.
    :param financial_command_kwargs: kwargs for financial_data_command
    :return:
    """
    print('Creating XLSX files with commands')
    create_all_xlsx_with_financials_commands(
        folder,
        company_id_list=company_ids,
        data_items_dict=data_items,
        **financial_command_kwargs
    )

    print('Populating XLSX files with Capital IQ data')
    populate_all_files_in_folder(
        folder,
        restart=restart,
        timeout=timeout,
        run_failed=run_failed
    )

    print('Combining individual XLSX files into a single CSV file')
    combine_all_capiq_xlsx(
        folder,
        outpath,
        restart=restart
    )


# TODO: download_ids, download_holdings, or combine those into the above function


import os
import string
import itertools
import math
import pandas as pd
from openpyxl.utils.dataframe import dataframe_to_rows

from exceldriver.workbook.create import get_workbook_and_worksheet
from .commands import financial_data_command, id_command, name_command, holdings_command

def create_all_xlsx_with_financials_commands(folder, company_id_list, data_items_dict, **financials_kwargs):
    [
        create_xlsx_with_financials_commands(folder, company_id, data_items_dict, **financials_kwargs)
        for company_id in company_id_list
    ]


def create_xlsx_with_financials_commands(folder, company_id, data_items_dict, **financials_kwargs):
    wb, ws = get_workbook_and_worksheet()
    _fill_with_financials_commands(ws, company_id, data_items_dict, **financials_kwargs)

    if not os.path.exists(folder):
        os.makedirs(folder)

    filepath = os.path.join(folder, f'{company_id}.xlsx')
    wb.save(filepath)

    return os.path.abspath(filepath)

def create_all_xlsx_with_holdings_commands(folder, company_id_list, date_str_list, data_items_dict):
    [
        create_xlsx_with_holdings_commands(folder, company_id, date_str, data_items_dict)
        for company_id, date_str in itertools.product(company_id_list, date_str_list)
    ]


def create_xlsx_with_holdings_commands(folder, company_id, date_str, data_items_dict):
    wb, ws = get_workbook_and_worksheet()
    _fill_with_holdings_commands(ws, company_id, date_str, data_items_dict)

    filepath = os.path.join(folder, f'{company_id} {_date_str_to_file_format(date_str)}.xlsx')
    wb.save(filepath)

    return os.path.abspath(filepath)

def create_all_xlsx_with_id_commands(inpath, folder, num_files=100):
    wb, ws = get_workbook_and_worksheet()

    df = pd.read_csv(inpath)
    _fill_id_column(df)
    _fill_capiq_id_column(df)
    _fill_capiq_name_column(df)

    rows_per_df = math.ceil(len(df)/num_files)

    count_per_wb = 0
    count_of_wb = 0
    for index, r in enumerate(dataframe_to_rows(df, index=False, header=True)):
        if index == 0:
            headers = r
        count_per_wb += 1
        ws.append(r)
        if count_per_wb >= rows_per_df:
            count_per_wb = 0
            count_of_wb += 1
            wb, ws = _save_wb_by_index_get_new_wb(count_of_wb, folder, wb)
            ws.append(headers)

##### Helper functions ####

def _save_wb_by_index_get_new_wb(index, folder, wb):
    filename = f'ids {index}.xlsx'
    filepath = os.path.join(folder, filename)
    wb.save(filepath)
    wb, ws = get_workbook_and_worksheet()
    return wb, ws

def _fill_id_column(df):
    """
    NOTE: inplace
    """
    # Fill id variable with ISIN, then if ISIN is missing, CUSIP
    df['ID'] = df['ISIN']
    missing_isin_mask = (pd.isnull(df['ISIN']))
    df.loc[missing_isin_mask, 'ID'] = df.loc[missing_isin_mask, 'CUSIP']

def _fill_capiq_id_column(df):
    """
    NOTE: inplace
    """
    df['Blank 1'] = df['ID'].apply(id_command)

    # Blank needed because ids will populate to the right by one column
    df['IQID'] = ''


def _fill_capiq_name_column(df):
    """
    NOTE: inplace
    """
    df['Blank 2'] = df['ID'].apply(name_command)

    # Blank needed because ids will populate to the right by one column
    df['IQ Name'] = ''

def _fill_with_financials_commands(ws, company_id, data_items_dict, **financials_kwargs):
    """
    Note: inplace
    """

    # Set default freq
    try:
        freq = financials_kwargs['freq']
    except KeyError:
        freq = 'Q'

    date_var, date_var_label = _get_date_var_and_label_from_freq(freq)

    column_generator = excel_cols()

    # Fill dates first
    current_column = next(column_generator)
    ws[f'{current_column}1'] = financial_data_command(company_id, date_var, **financials_kwargs, data_item_label=date_var_label)

    for item in data_items_dict:
        current_column = next(column_generator)
        ws[f'{current_column}1'] = financial_data_command(company_id, item, **financials_kwargs, data_item_label=data_items_dict[item])

def _fill_with_holdings_commands(ws, company_id, date_str, data_items_dict):
    column_generator = excel_cols()

    for item in data_items_dict:
        current_column = next(column_generator)
        ws[f'{current_column}1'] = holdings_command(
            company_id, item, date_str,
            data_item_label=data_items_dict[item]
        )

def _get_date_var_and_label_from_freq(freq):
    if freq == 'Y':
        date_var = (
            'IQ_FISCAL_Y', 'Fiscal Year'
        )
    elif freq == 'Q':
        date_var = (
            'IQ_ABS_PERIOD', 'Fiscal Quarter'
        )
    else:
        raise ValueError('Must pass Y or Q for freq')

    return date_var

def _date_str_to_file_format(date_str):
    return date_str.replace('/','-')

def excel_cols():
    n = 1
    while True:
        yield from (''.join(group) for group in itertools.product(string.ascii_uppercase, repeat=n))
        n += 1


import os
import re
import tempfile
import math
import traceback

import pandas as pd

from capiq_excel.tools.ext_pandas import _get_outpath_and_df_of_headers, _append_df_to_csv, append_csv_to_csv
from processfiles.files import FileProcessTracker


def combine_all_capiq_xlsx(infolder, outpath, restart=True, num_parts=100):

    file_tracker = FileProcessTracker(folder=infolder, restart=restart, file_types=('xlsx',))
    outpath, df_of_headers = _get_outpath_and_df_of_headers(outpath)
    all_columns = [col for col in df_of_headers.columns]

    # TODO: cleanup
    # Set up appending to many files to speed up process. Then the part files will be combined at the end
    num_files_per_part = math.ceil(len(file_tracker.process_list) / num_parts)
    file_num = 0
    with tempfile.TemporaryDirectory() as temp_dir:
        print(f'Creating temporary directory {temp_dir}')
        print(f'Running first pass of append. Will create {num_parts} files to be used in the final append.')
        for i, file in enumerate(file_tracker.file_generator()):
            # Every time we process num_files_per_part number of files, increment the output file
            if i % num_files_per_part == 0:
                file_num += 1
            temp_outpath = os.path.join(temp_dir, f'{file_num}.csv')
            df_of_headers, all_columns = _append_capiq_xlsx_to_csv(file, temp_outpath, df_of_headers, all_columns)

        # Now append created parts to output file
        print('Running second pass of append. Using in part files to create output file.')
        file_tracker = FileProcessTracker(folder=temp_dir, restart=True, file_types=('csv',))
        outpath, df_of_headers = _get_outpath_and_df_of_headers(outpath)
        all_columns = [col for col in df_of_headers.columns]
        for file in file_tracker.file_generator():
            df_for_append = pd.read_csv(file)  # load new data
            df_of_headers, all_columns = _append_df_to_csv(df_for_append, df_of_headers, outpath, all_columns)

def _append_capiq_xlsx_to_csv(file, outpath, df_of_headers, all_columns):
    df_for_append = pd.read_excel(file)  # load new data
    if _filepath_has_date(file):
        id_, date = _capiq_filepath_to_iq_id_and_date(file)
        df_for_append['CQID'] = id_
        df_for_append['Date'] = date
    else:
        df_for_append['CQID'] = _capiq_filepath_to_iq_id(file)
    df_of_headers, all_columns = _append_df_to_csv(df_for_append, df_of_headers, outpath, all_columns)

    return df_of_headers, all_columns

def _filepath_has_date(filepath):
    filename = os.path.basename(filepath)  # strips folders, etc.
    pattern = re.compile(r'(IQ\d+) ([\d-]+)([.]xlsx)')
    return True if pattern.match(filename) else False

def _capiq_filepath_to_iq_id(filepath):
    filename = os.path.basename(filepath) #strips folders, etc.
    pattern = re.compile(r'(IQ\d+)([.]xlsx)')
    return pattern.match(filename).group(1)

def _capiq_filepath_to_iq_id_and_date(filepath):
    filename = os.path.basename(filepath)  # strips folders, etc.
    pattern = re.compile(r'(IQ\d+) ([\d-]+)([.]xlsx)')
    match = pattern.match(filename)
    return match.group(1), match.group(2)
